__author__ = 'Jeffrey Knockel'
__copyright__ = 'Copyright 2013 Jeffrey Knockel'
__license__ = 'GPL v2'

from collections import deque
from copy import deepcopy
import os
from select import select
import signal
import socket
import struct
import threading

from dpkt import Packet, in_cksum
from dpkt.ip import IP, IP_MF
from dpkt.ip6 import IP6
from dpkt import tcp

from timeout import Timeout

ETH_P_IP = 0x800
ETH_P_IPV6 = 0x86dd
SO_BINDTODEVICE = 25

FLAG_TO_CHAR = [
    (tcp.TH_FIN, 'F'),
    (tcp.TH_SYN, 'S'),
    (tcp.TH_RST, 'R'),
    (tcp.TH_PUSH, 'P'),
    (tcp.TH_ACK, 'A'),
    (tcp.TH_URG, 'U'),
    (tcp.TH_ECE, 'E'),
    (tcp.TH_CWR, 'C'),
]

CHAR_TO_FLAG = dict((c, f) for f, c in FLAG_TO_CHAR)

def tcp_flags_to_string(n):
    s = ''
    for f, c in FLAG_TO_CHAR:
        if n & f:
            s += c
    return s

def tcp_string_to_flags(s):
    n = 0
    for c in s:
        n |= CHAR_TO_FLAG[c]
    return n

def fragment(p, fragsize=1480, firstsize=None):
    if type(p) is not IP:
        raise ValueError('p must be IP packet')
    if firstsize is None:
        firstsize = fragsize
    fragsize = (fragsize + 7) // 8 * 8
    firstsize = (firstsize + 7) // 8 * 8
    p = deepcopy(p)
    str(p) # calculate correct checksums
    s = str(p.data)
    p.data = ''
    ihl = len(str(p)) // 4
    nb = 1 + max((len(s) + fragsize - firstsize - 1) // fragsize, 0)
    lst = [None] * nb
    p.sum = 0
    for i in xrange(nb):
        q = deepcopy(p)
        if i == nb - 1:
            q.off &= ~IP_MF
        else:
            q.off |= IP_MF
        offset = 0 if i == 0 else firstsize + fragsize * (i - 1)
        size = firstsize if i == 0 else fragsize
        buf = s[offset:offset + size]
        q.len = ihl * 4 + len(buf)
        q.off = (q.off & ~0x1fff) | (offset // 8 & 0x1fff)
        q.data = buf
        lst[i] = q
    return lst

def fragment6(p, fragsize=1280, firstsize=None):
    if type(p) is not IP6:
        raise ValueError('p must be IP6 packet')
    if p.extension_hdrs[44] is None:
        raise ValueError('p must contain IP6FragmentHeader')
    p = deepcopy(p)
    if firstsize is None:
        firstsize = fragsize
    fragsize = (fragsize + 7) // 8 * 8
    firstsize = (firstsize + 7) // 8 * 8
    str(p) # calculate correct checksums
    s = str(p.data)
    p.p = 0
    p.data = ''
    pre = len(str(p))
    fragsize -= pre
    firstsize -= pre
    nb = 1 + max((len(s) + fragsize - firstsize - 1) // fragsize, 0)
    lst = [None] * nb
    for i in xrange(nb):
        q = deepcopy(p)
        frag = q.extension_hdrs[44]
        if i == nb - 1:
            frag.frag_off_resv_m &= ~1
        else:
            frag.frag_off_resv_m |= 1
        offset = 0 if i == 0 else firstsize + fragsize * (i - 1)
        size = firstsize if i == 0 else fragsize
        buf = s[offset:offset + size]
        q.plen = pre - 40 + len(buf)
        frag.frag_off_resv_m = (frag.frag_off_resv_m & ~0xfff8) | offset
        q.data = buf
        lst[i] = q
    return lst

def fragmentbuf(p, fragsize=1480, firstsize=None, copy=True):
    if type(p) is not IP:
        raise ValueError('p must be IP packet')
    if firstsize is None:
        firstsize = fragsize
    fragsize = (fragsize + 7) // 8 * 8
    firstsize = (firstsize + 7) // 8 * 8
    if copy:
        p = deepcopy(p)
    str(p) # calculate correct checksums
    s = str(p.data)
    old_data = p.data
    p.data = ''
    try:
        pbuf = str(p)
        ihl = len(pbuf) // 4
        off = p.off
        pbuf = [chr(0x40 | ihl), pbuf[1:2], None, pbuf[4:6], None, pbuf[8:10], None, pbuf[12:]]
        nb = 1 + max((len(s) + fragsize - firstsize - 1) // fragsize, 0)
        lst = [None] * nb
        for i in xrange(nb):
            if i == nb - 1:
                off &= ~IP_MF
            else:
                off |= IP_MF
            offset = 0 if i == 0 else firstsize + fragsize * (i - 1)
            size = firstsize if i == 0 else fragsize
            buf = s[offset:offset + size]
            pbuf[2] = struct.pack('!H', ihl * 4 + len(buf))
            pbuf[4] = struct.pack('!H', (off & ~0x1fff) | (offset // 8 & 0x1fff))
            pbuf[6] = '\0' * 2
            pbuf[6] = struct.pack('!H', in_cksum(''.join(pbuf)))
            lst[i] = ''.join(pbuf) + buf
        return lst
    finally:
        p.data = old_data

def fragmentbuf6(p, fragsize=1280, firstsize=None, copy=True):
    if type(p) is not IP6:
        raise ValueError('p must be IP6 packet')
    if p.extension_hdrs[44] is None:
        raise ValueError('p must contain IP6FragmentHeader')
    if copy:
        p = deepcopy(p)
    frag = p.extension_hdrs[44]
    if firstsize is None:
        firstsize = fragsize
    fragsize = (fragsize + 7) // 8 * 8
    firstsize = (firstsize + 7) // 8 * 8
    str(p) # calculate correct checksums
    s = str(p.data)
    old_p, old_data = p.p, p.data
    p.p, p.data = 0, ''
    try:
        pbuf = str(p)
        pre = len(pbuf)
        fragsize -= pre
        firstsize -= pre
        frag_off_resv_m = frag.frag_off_resv_m
        pbuf = [pbuf[:4], None, pbuf[6:pre - 6], None, pbuf[-4:]]
        nb = 1 + max((len(s) + fragsize - firstsize - 1) // fragsize, 0)
        lst = [None] * nb
        for i in xrange(nb):
            if i == nb - 1:
                frag_off_resv_m &= ~1
            else:
                frag_off_resv_m |= 1
            offset = 0 if i == 0 else firstsize + fragsize * (i - 1)
            size = firstsize if i == 0 else fragsize
            buf = s[offset:offset + size]
            pbuf[1] = struct.pack('!H', pre - 40 + len(buf))
            pbuf[3] = struct.pack('!H', (frag_off_resv_m & ~0xfff8) | offset)
            lst[i] = ''.join(pbuf) + buf
        return lst
    finally:
        p.p, p.data = old_p, old_data

def assert_main_thread():
    assert(threading.current_thread().name == 'MainThread')

def try_f(f, *args, **kwargs):
    try:
        return f(*args, **kwargs)
    except Exception:
        return None

class InterruptedException(Exception):
    pass

class BlockingTimeout(Timeout):
    def __init__(self, timeout=0.0, event=None):
        super(BlockingTimeout, self).__init__(timeout=timeout)
        self.event = threading.Event() if event is None else event
    def sleep(self):
        if self.event.wait(self.timeout()):
            raise InterruptedException()
    def interrupt(self):
        self.event.set()

class Queue(object):
    def __init__(self):
        self.condition = threading.Condition()
        self.queue = deque()
        self.interrupted = False
    def interrupt(self):
        with self.condition:
            self.interrupted = True
            self.queue.clear()
            self.condition.notify_all()
    def put(self, item):
        with self.condition:
            if not self.interrupted:
                self.queue.append(item)
                self.condition.notify()
    def get(self, timeout):
        with self.condition:
            while True:
                if self.interrupted:
                    raise InterruptedException()
                if len(self.queue) > 0:
                    return self.queue.popleft()
                remaining = timeout.timeout()
                if remaining <= 0.0:
                    return None
                self.condition.wait(remaining)

class SnifferSubscription(object):
    def __init__(self, sniffer):
        self.sniffer = sniffer
    def __enter__(self):
        sniffer = self.sniffer
        return sniffer
    def __exit__(self, type, value, traceback):
        self.sniffer.unsubscribe()

class Sniffer(object):
    def __init__(self, ip6=False, subscribe=False, iface=None):
        self.timeout_event = threading.Event()
        self.local = threading.local()
        self.local.callback = None
        self.lock = threading.Lock()
        self.queues = {}
        self.queues_lock = threading.Lock()
        self.interrupted = False
        self.thread = None
        if ip6:
            self.domain = socket.AF_INET6
            self.type = ETH_P_IPV6
            self.cls = IP6
            self.ntop = lambda n: socket.inet_ntop(socket.AF_INET6, n)
        else:
            self.domain = socket.AF_INET
            self.type = ETH_P_IP
            self.cls = IP
            self.ntop = lambda n: socket.inet_ntop(socket.AF_INET, n)
        self.iface = iface
        if subscribe:
            self.subscribe()
    def _sniff_socket(self):
        while True:
            rlist, wlist, xlist = select([self.ins, self.rpipe], [], [])
            if self.rpipe in rlist:
                break
            elif self.ins in rlist:
                buf, sa = self.ins.recvfrom(0xffff)
                with self.queues_lock:
                    for queue in self.queues.itervalues():
                        queue.put(buf)
    def _init(self):
        try:
            self.outs = socket.socket(self.domain, socket.SOCK_RAW, socket.IPPROTO_RAW)
            if self.iface is not None:
                self.outs.setsockopt(socket.SOL_SOCKET, SO_BINDTODEVICE, self.iface + '\0')
            self.ins = socket.socket(socket.AF_PACKET, socket.SOCK_DGRAM, socket.htons(self.type))
            if self.iface is not None:
                self.ins.bind((self.iface, self.type))
            self.rpipe, self.wpipe = os.pipe()
            self.thread = threading.Thread(target=self._sniff_socket)
            self.thread.start()
        except:
            self._uninit()
            raise
    def _uninit(self):
        if threading.current_thread() != self.thread:
            try_f(lambda: os.write(self.wpipe, 'x'))
            try_f(lambda: self.thread.join())
        try_f(lambda: os.close(self.wpipe))
        try_f(lambda: os.close(self.rpipe))
        try_f(lambda: self.ins.close())
        try_f(lambda: self.outs.close())
    def start_thread(self, target, args, kwargs):
        def wrapper():
            try:
                with self.subscribe():
                    target(self, *args, **kwargs)
            except InterruptedException:
                pass
        t = threading.Thread(target=wrapper)
        t.start()
        return t
    def subscribe(self):
        t = threading.current_thread()
        if t not in self.queues:
            with self.lock:
                if self.interrupted:
                    raise InterruptedException()
                with self.queues_lock:
                    self.queues[t] = Queue()
                    init = len(self.queues) == 1
                if init:
                    self._init()
        return SnifferSubscription(self)
    def unsubscribe(self):
        t = threading.current_thread()
        if t in self.queues:
            with self.lock:
                with self.queues_lock:
                    del self.queues[t]
                    uninit = len(self.queues) == 0
                if uninit:
                    self._uninit()
    def __enter__(self):
        return self
    def __exit__(self, type, value, traceback):
        self.shutdown()
    def sniff(self, timeout):
        queue = self.queues[threading.current_thread()]
        while True:
            buf = queue.get(timeout=timeout)
            if buf is None:
                return None
            return try_f(lambda: self.cls(buf))
    def send(self, ps):
        if isinstance(ps, Packet):
            ps = [ps]
        for p in ps:
            self.sendbuf(str(p), p.dst)
    def sendbuf(self, bufs, dst):
        if isinstance(bufs, str):
            bufs = [bufs]
        outs = self.outs
        for buf in bufs:
            outs.sendto(buf, (self.ntop(dst), 0))
    def shutdown(self):
        self.unsubscribe()
        with self.lock:
            self.interrupted = True
            self.timeout_event.set()
            with self.queues_lock:
                for queue in self.queues.itervalues():
                    queue.interrupt()
        try_f(lambda: self.thread.join())
    def timeout(self, timeout=0.0):
        return BlockingTimeout(timeout=timeout, event=self.timeout_event)
    def set_callback(self, callback):
        self.local.callback = callback
    def work(self):
        if self.local.callback is not None:
            self.local.callback()

def start_sniffing(target, ip6=False, args=None, kwargs=None):
    assert_main_thread()
    old_handler = signal.getsignal(signal.SIGINT)
    assert(old_handler == signal.default_int_handler)
    interrupted = {}
    def interrupt_ignore(signum, frame):
        interrupted['signum'] = signum
        interrupted['frame'] = frame
        signal.signal(signal.SIGINT, signal.SIG_IGN)
    def interrupt_raise(signum, frame):
        interrupt_ignore(signum, frame)
        raise KeyboardInterrupt()
    sniffer = None
    try:
        signal.signal(signal.SIGINT, interrupt_ignore)
        sniffer = Sniffer(ip6=ip6, subscribe=True)
        signal.signal(signal.SIGINT, interrupt_raise)
        if interrupted:
            raise KeyboardInterrupt()
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        result = target(sniffer, *args, **kwargs)
        signal.signal(signal.SIGINT, interrupt_ignore)
        return result
    except:
        signal.signal(signal.SIGINT, interrupt_ignore)
        raise
    finally:
        try_f(lambda: sniffer.shutdown())
        try_f(lambda: signal.signal(signal.SIGINT, old_handler))
        if interrupted:
            raise KeyboardInterrupt()

def start_sniffing4(target, *args, **kwargs):
    return start_sniffing(target, ip6=False, args=args, kwargs=kwargs)

def start_sniffing6(target, *args, **kwargs):
    return start_sniffing(target, ip6=True, args=args, kwargs=kwargs)
